@@ -1,4 +1,5 @@
 """CLI tool for sampling sentences from intents."""
+
 import argparse
 import itertools
 import json
@@ -9,22 +10,28 @@ from pathlib import Path
 from typing import Dict, Iterable, Optional, Set, Tuple
 
 import yaml
+from unicode_rbnf import RbnfEngine
 
+from .errors import MissingListError, MissingRuleError
 from .expression import (
+    Alternative,
     Expression,
+    Group,
     ListReference,
+    Permutation,
     RuleReference,
     Sentence,
     Sequence,
-    SequenceType,
     TextChunk,
 )
-from .intents import Intents, RangeSlotList, SlotList, TextSlotList
-from .recognize import MissingListError, MissingRuleError
+from .intents import Intents, RangeSlotList, SlotList, TextSlotList, WildcardSlotList
 from .util import merge_dict, normalize_whitespace
 
 _LOGGER = logging.getLogger("hassil.sample")
 
+# lang -> engine
+_ENGINE_CACHE: Dict[str, RbnfEngine] = {}
+
 
 def sample_intents(
     intents: Intents,
@@ -32,6 +39,9 @@ def sample_intents(
     expansion_rules: Optional[Dict[str, Sentence]] = None,
     max_sentences_per_intent: Optional[int] = None,
     intent_names: Optional[Set[str]] = None,
+    language: Optional[str] = None,
+    exclude_sentences_with_wildcards: bool = True,
+    expand_ranges: bool = True,
 ) -> Iterable[Tuple[str, str]]:
     """Sample text strings for sentences from intents."""
     if slot_lists is None:
@@ -58,11 +68,27 @@ def sample_intents(
         skip_intent = False
 
         for intent_data in intent.data:
+            if intent_data.expansion_rules:
+                local_expansion_rules = {
+                    **expansion_rules,
+                    **intent_data.expansion_rules,
+                }
+            else:
+                local_expansion_rules = expansion_rules
+
             for intent_sentence in intent_data.sentences:
-                sentence_texts = sample_expression(
+                if exclude_sentences_with_wildcards and any(
+                    list_name in intent_data.wildcard_list_names
+                    for list_name in intent_sentence.list_names(local_expansion_rules)
+                ):
+                    continue
+
+                sentence_texts = sample_sentence(
                     intent_sentence,
                     slot_lists,
-                    expansion_rules,
+                    local_expansion_rules,
+                    language=language,
+                    expand_ranges=expand_ranges,
                 )
                 for sentence_text in sentence_texts:
                     yield (intent_name, sentence_text)
@@ -81,41 +107,93 @@ def sample_intents(
                 break
 
 
+def sample_sentence(
+    sentence: Sentence,
+    slot_lists: Optional[Dict[str, SlotList]] = None,
+    expansion_rules: Optional[Dict[str, Sentence]] = None,
+    language: Optional[str] = None,
+    expand_lists: bool = True,
+    expand_ranges: bool = True,
+) -> Iterable[str]:
+    return sample_expression(
+        sentence.expression,
+        slot_lists,
+        expansion_rules,
+        language,
+        expand_lists,
+        expand_ranges,
+    )
+
+
 def sample_expression(
     expression: Expression,
     slot_lists: Optional[Dict[str, SlotList]] = None,
     expansion_rules: Optional[Dict[str, Sentence]] = None,
+    language: Optional[str] = None,
+    expand_lists: bool = True,
+    expand_ranges: bool = True,
 ) -> Iterable[str]:
     """Sample possible text strings from an expression."""
     if isinstance(expression, TextChunk):
         chunk: TextChunk = expression
         yield chunk.original_text
-    elif isinstance(expression, Sequence):
-        seq: Sequence = expression
-        if seq.type == SequenceType.ALTERNATIVE:
-            for item in seq.items:
+    elif isinstance(expression, Group):
+        grp: Group = expression
+        if isinstance(grp, Alternative):
+            for item in grp.items:
                 yield from sample_expression(
                     item,
                     slot_lists,
                     expansion_rules,
+                    language=language,
+                    expand_lists=expand_lists,
+                    expand_ranges=expand_ranges,
                 )
-        elif seq.type == SequenceType.GROUP:
+        elif isinstance(grp, Sequence):
             seq_sentences = map(
                 partial(
                     sample_expression,
                     slot_lists=slot_lists,
                     expansion_rules=expansion_rules,
+                    language=language,
+                    expand_lists=expand_lists,
+                    expand_ranges=expand_ranges,
                 ),
-                seq.items,
+                grp.items,
             )
             sentence_texts = itertools.product(*seq_sentences)
             for sentence_words in sentence_texts:
                 yield normalize_whitespace("".join(sentence_words))
+        elif isinstance(grp, Permutation):
+            # Need to make lists instead because itertools does multiple passes.
+            grp_sentences = [
+                list(
+                    sample_expression(
+                        item,
+                        slot_lists,
+                        expansion_rules,
+                        language=language,
+                        expand_lists=expand_lists,
+                        expand_ranges=expand_ranges,
+                    )
+                )
+                for item in grp.items
+            ]
+            for perm_sentences in itertools.permutations(grp_sentences):
+                sentence_texts = itertools.product(*perm_sentences)
+                for sentence_words in sentence_texts:
+                    # Strip added whitespace
+                    yield normalize_whitespace("".join(sentence_words)).strip()
         else:
-            raise ValueError(f"Unexpected sequence type: {seq}")
+            raise ValueError(f"Unexpected group type: {grp}")
     elif isinstance(expression, ListReference):
         # {list}
         list_ref: ListReference = expression
+
+        if not expand_lists:
+            yield f"{{{list_ref.list_name}}}"
+            return
+
         if (not slot_lists) or (list_ref.list_name not in slot_lists):
             raise MissingListError(f"Missing slot list {{{list_ref.list_name}}}")
 
@@ -132,14 +210,52 @@ def sample_expression(
                     text_value.text_in,
                     slot_lists,
                     expansion_rules,
+                    language=language,
+                    expand_lists=expand_lists,
+                    expand_ranges=expand_ranges,
                 )
         elif isinstance(slot_list, RangeSlotList):
             range_list: RangeSlotList = slot_list
-            number_strs = map(
-                str, range(range_list.start, range_list.stop + 1, range_list.step)
-            )
-            yield from number_strs
 
+            if not expand_ranges:
+                if range_list.name:
+                    yield f"{{{range_list.name}}}"
+                else:
+                    yield "{number}"
+                return
+
+            if range_list.digits:
+                number_strs = map(str, range_list.get_numbers())
+                yield from number_strs
+
+            if range_list.words:
+                words_language = range_list.words_language or language
+                if words_language:
+                    engine = _ENGINE_CACHE.get(words_language)
+                    if engine is None:
+                        engine = RbnfEngine.for_language(words_language)
+                        _ENGINE_CACHE[words_language] = engine
+
+                    assert engine is not None
+
+                    # digits -> words
+                    for word_number in range_list.get_numbers():
+                        # Use all unique words for a number, including different
+                        # genders, cases, etc.
+                        format_result = engine.format_number(word_number)
+                        unique_number_strs = set(format_result.text_by_ruleset.values())
+                        yield from unique_number_strs
+                else:
+                    _LOGGER.warning(
+                        "No language set, so cannot convert %s digits to words",
+                        list_ref.slot_name,
+                    )
+        elif isinstance(slot_list, WildcardSlotList):
+            wildcard_list: WildcardSlotList = slot_list
+            if wildcard_list.name:
+                yield f"{{{wildcard_list.name}}}"
+            else:
+                yield "{wildcard}"
         else:
             raise ValueError(f"Unexpected slot list type: {slot_list}")
     elif isinstance(expression, RuleReference):
@@ -148,11 +264,14 @@ def sample_expression(
         if (not expansion_rules) or (rule_ref.rule_name not in expansion_rules):
             raise MissingRuleError(f"Missing expansion rule <{rule_ref.rule_name}>")
 
-        rule_body = expansion_rules[rule_ref.rule_name]
+        rule_body = expansion_rules[rule_ref.rule_name].expression
         yield from sample_expression(
             rule_body,
             slot_lists,
             expansion_rules,
+            language=language,
+            expand_lists=expand_lists,
+            expand_ranges=expand_ranges,
         )
     else:
         raise ValueError(f"Unexpected expression: {expression}")
@@ -180,6 +299,7 @@ def main():
     parser.add_argument(
         "--names", nargs="+", default=["entity"], help="Device/entity names"
     )
+    parser.add_argument("--language", help="Language for digits to words")
     parser.add_argument(
         "--debug", action="store_true", help="Print DEBUG messages to the console"
     )
@@ -215,6 +335,7 @@ def main():
         slot_lists,
         max_sentences_per_intent=args.max_sentences_per_intent,
         intent_names=set(args.intents) if args.intents else None,
+        language=args.language,
     )
     for intent_name, sentence_text in intents_and_texts:
         json.dump(
